@@ -67,10 +67,9 @@ typedef UWORD WORD;
 
 // 128 gm and 63 drum
 #define MAXSMP				191
-static int isabspath = 0;
-static char midipat[MAXSMP][128];
-static char pathforpat[128];
-static char timiditycfg[128];
+static char midipat[MAXSMP][PATH_MAX];
+static char pathforpat[PATH_MAX] = {};
+static char timiditycfg[PATH_MAX] = {};
 
 #pragma pack(1)
 
@@ -323,6 +322,7 @@ typedef struct {
 	char *mm;
 	int sz;
 	int pos;
+	int error;
 } MMFILE;
 
 static long mmftell(MMFILE *mmfile)
@@ -332,111 +332,156 @@ static long mmftell(MMFILE *mmfile)
 
 static void mmfseek(MMFILE *mmfile, long p, int whence)
 {
+	int newpos = mmfile->pos;
 	switch(whence) {
 		case SEEK_SET:
-			mmfile->pos = p;
+			newpos = p;
 			break;
 		case SEEK_CUR:
-			mmfile->pos += p;
+			newpos += p;
 			break;
 		case SEEK_END:
-			mmfile->pos = mmfile->sz + p;
+			newpos = mmfile->sz + p;
 			break;
 	}
+	if (newpos < mmfile->sz)
+		mmfile->pos = newpos;
+	else {
+		mmfile->error = 1;
+//		printf("WARNING: seeking too far\n");
+	}
 }
 
 static void mmreadUBYTES(BYTE *buf, long sz, MMFILE *mmfile)
 {
-	memcpy(buf, &mmfile->mm[mmfile->pos], sz);
+	int sztr = sz;
+	// do not overread.
+	if (sz > mmfile->sz - mmfile->pos)
+		sztr = mmfile->sz - mmfile->pos;
+	memcpy(buf, &mmfile->mm[mmfile->pos], sztr);
 	mmfile->pos += sz;
+	// if truncated read, populate the rest of the array with zeros.
+	if (sz > sztr)
+		memset(buf+sztr, 0, sz-sztr);
 }
 
 static void mmreadSBYTES(char *buf, long sz, MMFILE *mmfile)
 {
+	// do not overread.
+	if (sz > mmfile->sz - mmfile->pos)
+		sz = mmfile->sz - mmfile->pos;
 	memcpy(buf, &mmfile->mm[mmfile->pos], sz);
 	mmfile->pos += sz;
 }
 
 #endif
 
+long _mm_getfsize(MMSTREAM *mmpat) {
+	long fsize;
+	_mm_fseek(mmpat, 0L, SEEK_END);
+	fsize = _mm_ftell(mmpat);
+	_mm_fseek(mmpat, 0L, SEEK_SET);
+	return(fsize);
+}
+
 void pat_init_patnames(void)
 {
-	int i,j, pfnlen, ndir;
+	int z, i, nsources, isdrumset, nskip, pfnlen;
 	char *p, *q;
 	char line[PATH_MAX];
+	char cfgsources[5][PATH_MAX] = {{0}, {0}, {0}, {0}, {0}};
 	MMSTREAM *mmcfg;
-	strcpy(pathforpat, PATHFORPAT);
-	strcpy(timiditycfg, TIMIDITYCFG);
+	strncpy(pathforpat, PATHFORPAT, PATH_MAX);
+	strncpy(timiditycfg, TIMIDITYCFG, PATH_MAX);
 	p = getenv(PAT_ENV_PATH2CFG);
 	if( p ) {
-		strcpy(timiditycfg,p);
-		strcpy(pathforpat,p);
-		strcat(timiditycfg,"/timidity.cfg");
-		strcat(pathforpat,"/instruments");
+		strncpy(timiditycfg, p, PATH_MAX - 14);
+		strncpy(pathforpat, p, PATH_MAX - 13);
+		strcat(timiditycfg, "/timidity.cfg");
+		strcat(pathforpat, "/instruments");
 	}
-	mmcfg = _mm_fopen(timiditycfg,"r");
+	strncpy(cfgsources[0], timiditycfg, PATH_MAX - 1);
+	nsources = 1;
+
 	for( i=0; i<MAXSMP; i++ )	midipat[i][0] = '\0';
-	if( !mmcfg ) {
-		pat_message("can not open %s, use environment variable " PAT_ENV_PATH2CFG " for the directory", timiditycfg);
-	}
-	else {
-		// read in bank 0 and drum patches
-		j = 0;
-		_mm_fgets(mmcfg, line, PATH_MAX);
-		while( !_mm_feof(mmcfg) ) {
-			if( isdigit(line[0]) || (line[0] == '\t' && isdigit(line[1])) ) {
-				i = atoi(line);
-				if( i < MAXSMP && i >= 0 ) {
-					p = strchr(line,'/') + 1;
-					if(j) 
-						q = midipat[pat_gm_drumnr(i)-1];
-					else
-						q = midipat[i];
+
+	for ( z=0; z<5; z++ ) {
+		if (cfgsources[z][0] == 0) continue;
+		mmcfg = _mm_fopen(cfgsources[z],"r");
+		if( !mmcfg ) {
+			pat_message("can not open %s, use environment variable " PAT_ENV_PATH2CFG " for the directory", cfgsources[z]);
+		}
+		else {
+			// read in bank 0 and drum patches
+			isdrumset = 0;
+			_mm_fgets(mmcfg, line, PATH_MAX);
+			while( !_mm_feof(mmcfg) ) {
+			if( isdigit(line[0]) || (isblank(line[0]) && isdigit(line[1])) ) {
+				p = line;
+				// get pat number
+				while ( isspace(*p) ) p ++;
+				i = atoi(p);
+				while ( isdigit(*p) ) p ++;
+				while ( isspace(*p) ) p ++;
+				// get pat file name
+				if( *p && i < MAXSMP && i >= 0 && *p != '#' ) {
+					q = isdrumset ? midipat[pat_gm_drumnr(i)-1] : midipat[i];
 					pfnlen = 0;
-					ndir = 0;
-					while( *p && !isspace(*p) && pfnlen < 128 ) {
-						if (*p == DIRDELIM) ndir++;
+					while( *p && !isspace(*p) && *p != '#' && pfnlen < 128 ) {
 						pfnlen ++;
 						*q++ = *p++;
 					}
-					if (ndir > 2) isabspath = 1;
 					if( isblank(*p) && *(p+1) != '#' && pfnlen < 128 ) {
-						*q++ = ':';
+						*q++ = ':'; pfnlen ++;
 						while( isspace(*p) ) {
 							while( isspace(*p) ) p++;
 							if ( *p == '#' ) { // comment
-								
+
 							} else while( *p && !isspace(*p) && pfnlen < 128 ) {
 								pfnlen ++;
 								*q++ = *p++;
 							}
-							if( isspace(*p) ) *q++ = ' ';
+							if( isspace(*p) ) { *q++ = ' '; pfnlen++; }
 						}
 					}
 					*q++ = '\0';
 				}
 			}
-			if( !strncmp(line,"drumset",7) ) j = 1;
+			if( !strncmp(line,"drumset",7) ) isdrumset = 1;
+			if( !strncmp(line,"source",6) && nsources < 5 ) {
+				q = cfgsources[nsources];
+				p = &line[7];
+				while ( isspace(*p) ) p ++;
+				pfnlen = 0;
+				while ( *p && *p != '#' && !isspace(*p) && pfnlen < 128 ) {
+					pfnlen ++;
+					*q++ = *p++;
+				}
+				*q = 0; // null termination
+				nsources++;
+			}
 			_mm_fgets(mmcfg, line, PATH_MAX);
+
+			} /* end file parsing */
+			_mm_fclose(mmcfg);
 		}
-		_mm_fclose(mmcfg);
 	}
 	q = midipat[0];
-	j = 0;
-	// make all empty patches duplicate the previous valid one.
+	nskip = 0;
+	// make all empty patches duplicates the previous valid one.
 	for( i=0; i<MAXSMP; i++ )	{
 		if( midipat[i][0] ) q = midipat[i];
 		else {
 			if( midipat[i] != q)
-				strcpy(midipat[i],q);
-			if( midipat[i][0] == '\0' ) j++;
+				strcpy(midipat[i], q);
+			if( midipat[i][0] == '\0' ) nskip++;
 		}
 	}
-	if( j ) {
+	if( nskip ) {
 		for( i=MAXSMP; i-- > 0; )	{
 			if( midipat[i][0] ) q = midipat[i];
 			else if( midipat[i] != q )
-				strcpy(midipat[i],q);
+				strcpy(midipat[i], q);
 		}
 	}
 }
@@ -444,14 +489,16 @@ void pat_init_patnames(void)
 static char *pat_build_path(char *fname, int pat)
 {
 	char *ps;
-	ps = strrchr(midipat[pat], ':');
-	
+	char *patfile = midipat[pat];
+	int isabspath = (patfile[0] == '/');
+	if ( isabspath ) patfile ++;
+	ps = strrchr(patfile, ':');
 	if( ps ) {
-		sprintf(fname, "%s%c%s", isabspath ? "" : pathforpat, DIRDELIM, midipat[pat]);
+		sprintf(fname, "%s%c%s", isabspath ? "" : pathforpat, DIRDELIM, patfile);
 		strcpy(strrchr(fname, ':'), ".pat");
 		return ps;
 	}
-	sprintf(fname, "%s%c%s.pat", isabspath ? "" : pathforpat, DIRDELIM, midipat[pat]);
+	sprintf(fname, "%s%c%s.pat", isabspath ? "" : pathforpat, DIRDELIM, patfile);
 	return 0;
 }
 
@@ -519,7 +566,8 @@ static void pat_read_waveheader(MMSTREAM *mmpat, WaveHeader *hw, int layer)
 		}
 	}
 	_mm_read_UBYTES((BYTE *)hw, sizeof(WaveHeader), mmpat);
-	strncpy(hw->reserved, hl.reserved, 36);
+	strncpy(hw->reserved, hl.reserved, 32);
+	hw->reserved[31] = 0;
 	if( hw->start_loop >= hw->wave_size ) {
 		hw->start_loop = 0;
 		hw->end_loop = 0;
@@ -544,6 +592,10 @@ static void pat_get_waveheader(MMFILE *mmpat, WaveHeader *hw, int layer)
 			for( i=1; i<layer; i++ ) {
 				mmreadUBYTES((BYTE *)hw, sizeof(WaveHeader), mmpat);
 				mmfseek(mmpat, hw->wave_size, SEEK_CUR);
+				if ( mmpat->error ) {
+					hw->wave_size = 0;
+					return;
+				}
 			}
 		}
 		else {
@@ -578,13 +630,17 @@ static void pat_get_waveheader(MMFILE *mmpat, WaveHeader *hw, int layer)
 static int pat_readpat_attr(int pat, WaveHeader *hw, int layer)
 {
 	char fname[128];
+	unsigned long fsize;
 	MMSTREAM *mmpat;
 	pat_build_path(fname, pat);
 	mmpat = _mm_fopen(fname, "r");
 	if( !mmpat )
 		return 0;
+	fsize = _mm_getfsize(mmpat);
 	pat_read_waveheader(mmpat, hw, layer);
 	_mm_fclose(mmpat);
+	if (hw->wave_size > fsize)
+		return 0;
 	return 1;
 }
 
@@ -603,7 +659,7 @@ static void pat_amplify(char *b, int num, int amp, int m)
 			for( i=0; i<n; i++ ) {
 				v = (((int)(*pw) - 0x8000) * amp) / 100;
 				if( v < -0x8000 ) v = -0x8000;
-				if( v >  0x7fff ) v =  0x7fff; 
+				if( v >  0x7fff ) v =  0x7fff;
 				*pw++ = v + 0x8000;
 			}
 		}
@@ -612,7 +668,7 @@ static void pat_amplify(char *b, int num, int amp, int m)
 			for( i=0; i<n; i++ ) {
 				v = ((*pi) * amp) / 100;
 				if( v < -0x8000 ) v = -0x8000;
-				if( v >  0x7fff ) v =  0x7fff; 
+				if( v >  0x7fff ) v =  0x7fff;
 				*pi++ = v;
 			}
 		}
@@ -797,7 +853,7 @@ static PATHANDLE *PAT_Init(void)
     PATHANDLE   *retval;
 #ifdef NEWMIKMOD
     MM_ALLOC     *allochandle;
-    
+
 		allochandle = _mmalloc_create("Load_PAT", NULL);
     retval = (PATHANDLE *)_mm_calloc(allochandle, 1,sizeof(PATHANDLE));
 		if( !retval ) return NULL;
@@ -1057,7 +1113,7 @@ static void pat_setpat_inst(WaveHeader *hw, INSTRUMENT *d, int smp)
 {
 	int u, inuse;
 	int envpoint[6], envvolume[6];
-	for(u=0; u<120; u++) {        
+	for(u=0; u<120; u++) {
 		d->samplenumber[u] = smp;
 		d->samplenote[u] = smp;
 	}
@@ -1269,7 +1325,7 @@ static void PATsample(CSoundFile *cs, MODINSTRUMENT *q, int smp, int gm)
 
 		// Enable aggressive declicking for songs that do not loop and that
 		// are long enough that they won't be adversely affected.
-		
+
 		q->flags  |= SL_LOOP;
 		q->format |= SF_16BITS;
 		q->format |= SF_SIGNED;
@@ -1279,7 +1335,7 @@ static void PATsample(CSoundFile *cs, MODINSTRUMENT *q, int smp, int gm)
 #else
 	s[31] = '\0';
 	memset(cs->m_szNames[smp], 0, 32);
-	strcpy(cs->m_szNames[smp], s);
+	strncpy(cs->m_szNames[smp], s, 32-1);
 	q->nGlobalVol = 64;
 	q->nPan       = 128;
 	q->uFlags     = CHN_16BIT;
@@ -1288,7 +1344,7 @@ static void PATsample(CSoundFile *cs, MODINSTRUMENT *q, int smp, int gm)
 		pat_setpat_attr(&hw, q);
 		pat_loops[smp-1] = (q->uFlags & CHN_LOOP)? 1: 0;
 		if( hw.modes & PAT_16BIT ) p = (char *)malloc(hw.wave_size);
-		else p = (char *)malloc(hw.wave_size * sizeof(short int));
+		else p = (char *)malloc(hw.wave_size * sizeof(char)*2);
 		if( p ) {
 			if( hw.modes & PAT_16BIT ) {
 				dec_pat_Decompress16Bit((short int *)p, hw.wave_size>>1, gm - 1);
@@ -1310,7 +1366,7 @@ static void PATsample(CSoundFile *cs, MODINSTRUMENT *q, int smp, int gm)
 		q->nVolume    = 256;
 		q->uFlags    |= CHN_LOOP;
 		q->uFlags    |= CHN_16BIT;
-		p = (char *)malloc(q->nLength*sizeof(short int));
+		p = (char *)malloc(q->nLength*sizeof(char)*2);
 		if( p ) {
 			dec_pat_Decompress8Bit((short int *)p, q->nLength, smp + MAXSMP - 1);
 			cs->ReadSample(q, RS_PCM16S, (LPSTR)p, q->nLength*2);
@@ -1395,6 +1451,7 @@ BOOL CSoundFile::ReadPAT(const BYTE *lpStream, DWORD dwMemLength)
 	mm.mm = (char *)lpStream;
 	mm.sz = dwMemLength;
 	mm.pos = 0;
+	mm.error = 0;
 #endif
 	while( avoid_reentry ) sleep(1);
 	avoid_reentry = 1;
@@ -1470,12 +1527,12 @@ BOOL CSoundFile::ReadPAT(const BYTE *lpStream, DWORD dwMemLength)
 	}
 #else
 	m_nType         = MOD_TYPE_PAT;
-	m_nInstruments  = h->samples + 1; // we know better but use each sample in the pat...
-	m_nSamples      = h->samples + 1; // xmms modplug does not use slot zero
+	m_nInstruments  = h->samples >= MAX_INSTRUMENTS-1 ? MAX_INSTRUMENTS-1 : h->samples + 1; // we know better but use each sample in the pat...
+	m_nSamples      = h->samples >= MAX_SAMPLES-1 ? MAX_SAMPLES-1 : h->samples + 1; // xmms modplug does not use slot zero
 	m_nDefaultSpeed = 6;
 	m_nChannels     = h->samples;
 	numpat          = t;
-	
+
 	m_dwSongFlags   = SONG_LINEARSLIDES;
 	m_nMinPeriod    = 28 << 2;
 	m_nMaxPeriod    = 1712 << 3;
@@ -1523,8 +1580,9 @@ BOOL CSoundFile::ReadPAT(const BYTE *lpStream, DWORD dwMemLength)
 		s[31] = '\0';
 		memset(m_szNames[t], 0, 32);
 		strcpy(m_szNames[t], s);
-		if( hw.modes & PAT_16BIT ) p = (char *)malloc(hw.wave_size);
-		else p = (char *)malloc(hw.wave_size * sizeof(short int));
+		if ( hw.wave_size == 0 ) p = NULL;
+		else if( hw.modes & PAT_16BIT ) p = (char *)malloc(hw.wave_size);
+		else p = (char *)malloc(hw.wave_size * sizeof(char) * 2);
 		if( p ) {
 			mmreadSBYTES(p, hw.wave_size, mmfile);
 			if( hw.modes & PAT_16BIT ) {
