@@ -5,6 +5,7 @@
 */
 
 #include "stdafx.h"
+#include <stdlib.h>
 #include "sndfile.h"
 #include "tables.h"
 
@@ -18,7 +19,7 @@
 DWORD CSoundFile::GetLength(BOOL bAdjust, BOOL bTotal)
 //----------------------------------------------------
 {
-	UINT dwElapsedTime=0, nRow=0, nCurrentPattern=0, nNextPattern=0, nPattern=Order[0];
+	UINT dwElapsedTime=0, nRow=0, nCurrentPattern=0, nNextPattern=0, nPattern=0;
 	UINT nMusicSpeed=m_nDefaultSpeed, nMusicTempo=m_nDefaultTempo, nNextRow=0;
 	UINT nMaxRow = 0, nMaxPattern = 0;
 	LONG nGlbVol = m_nDefaultGlobalVolume, nOldGlbVolSlide = 0;
@@ -37,12 +38,10 @@ DWORD CSoundFile::GetLength(BOOL bAdjust, BOOL bTotal)
 	memset(oldparam, 0, sizeof(oldparam));
 	memset(chnvols, 64, sizeof(chnvols));
 	memset(samples, 0, sizeof(samples));
-	for (UINT icv=0; icv<m_nChannels; icv++) chnvols[icv] = ChnSettings[icv].nVolume;
+	for (UINT icv=0; icv<m_nChannels; icv++)
+		chnvols[icv] = ChnSettings[icv].nVolume;
 	nMaxRow = m_nNextRow;
 	nMaxPattern = m_nNextPattern;
-	nCurrentPattern = nNextPattern = 0;
-	nPattern = Order[0];
-	nRow = nNextRow = 0;
 	for (;;)
 	{
 		UINT nSpeedCount = 0;
@@ -1881,7 +1880,7 @@ void CSoundFile::ProcessMidiMacro(UINT nChn, LPCSTR pszMidiMacro, UINT param)
 		case '0':
 			{
 				int oldcutoff = pChn->nCutOff;
-				if (dwParam < 0x80) pChn->nCutOff = dwParam;
+				if (dwParam < 0x80) pChn->nCutOff = (BYTE)dwParam;
 #ifndef NO_FILTER
 				oldcutoff -= pChn->nCutOff;
 
@@ -1895,7 +1894,7 @@ void CSoundFile::ProcessMidiMacro(UINT nChn, LPCSTR pszMidiMacro, UINT param)
 
 		// F0.F0.01.xx: Set Resonance
 		case '1':
-			if (dwParam < 0x80) pChn->nResonance = dwParam;
+			if (dwParam < 0x80) pChn->nResonance = (BYTE)dwParam;
 #ifndef NO_FILTER
 			SetupChannelFilter(pChn, (pChn->dwFlags & CHN_FILTER) ? FALSE : TRUE);
 #endif // NO_FILTER
